from argparse import Namespace

from torch import nn
from torch.nn import ModuleList

from ml_lib.modules.blocks import ConvModule, LinearModule
from ml_lib.modules.util import (LightningBaseModule, Splitter, Merger)
from util.module_mixins import (BaseOptimizerMixin, BaseTrainMixin, BaseValMixin, DatasetMixin,
                                BaseDataloadersMixin)


class BandwiseConvClassifier(DatasetMixin,
                             BaseDataloadersMixin,
                             BaseTrainMixin,
                             BaseValMixin,
                             BaseOptimizerMixin,
                             LightningBaseModule
                             ):

    def __init__(self, hparams):
        super(BandwiseConvClassifier, self).__init__(hparams)

        # Dataset
        # =============================================================================
        self.dataset = self.build_dataset()

        # Model Paramters
        # =============================================================================
        # Additional parameters
        self.in_shape = self.dataset.train_dataset.sample_shape
        self.conv_filters = self.params.filters
        self.n_band_sections = 4

        # Modules
        # =============================================================================
        self.split = Splitter(self.in_shape, self.n_band_sections)

        k = 3
        self.band_list = ModuleList()
        for band in range(self.n_band_sections):
            last_shape = self.split.shape
            conv_list = ModuleList()
            for filters in self.conv_filters:
                conv_list.append(ConvModule(last_shape, filters, (k,k), conv_stride=(2, 2), conv_padding=2,
                                            **self.params.module_kwargs))
                last_shape = conv_list[-1].shape
                # self.conv_list.append(ConvModule(last_shape, 1, 1, conv_stride=1, **self.params.module_kwargs))
                # last_shape = self.conv_list[-1].shape
            self.band_list.append(conv_list)

        self.merge = Merger(self.band_list[-1][-1].shape, self.n_band_sections)

        self.full_1 = LinearModule(self.merge.shape, self.params.lat_dim, **self.params.module_kwargs)
        self.full_2 = LinearModule(self.full_1.shape, self.full_1.shape * 2, **self.params.module_kwargs)
        self.full_3 = LinearModule(self.full_2.shape, self.full_2.shape // 2, **self.params.module_kwargs)

        self.full_out = LinearModule(self.full_3.shape, 1, bias=self.params.bias, activation=nn.Sigmoid)

    def forward(self, batch, **kwargs):
        tensors = self.split(batch)
        for idx, (tensor, convs) in enumerate(zip(tensors, self.band_list)):
            for conv in convs:
                tensor = conv(tensor)
            tensors[idx] = tensor

        tensor = self.merge(tensors)
        tensor = self.full_1(tensor)
        tensor = self.full_2(tensor)
        tensor = self.full_3(tensor)
        tensor = self.full_out(tensor)
        return Namespace(main_out=tensor)
